/**
  *  \file client/dlg-receiver.cc
  *  \brief Message Receiver Widget
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-receiver.h"
#include "game/msg.h"
#include "ui/window.h"
#include "ui/layout.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "ui/uiglobal.h"
#include "gfx/palette.h"
#include "ui/widgets.h"
#include "ui/msgbox.h"
#include "game/playerset.h"
#include "game/racename.h"

/**
  *  \class WReceiverSelector
  *  \brief Message Receiver Selector
  *
  *  This widget displays a list of players, and lets the user choose
  *  an arbitrary combination.
  */

/** Constructor.
    \param id widget id */
WReceiverSelector::WReceiverSelector(int id)
    : UIWidget(id), mdown(false)
{ }

WReceiverSelector::~WReceiverSelector()
{ }

/** Toggle whether this is a universal message. When all players are
    selected, unselects them. Otherwise, selects them. */
void
WReceiverSelector::toggleUniversal()
{
    setSelectedWidget();
    if (receivers.contains(GPlayerSet::fromInteger(GOutbox::UNIVERSAL_MASK)))
        setReceivers(receivers - GPlayerSet::fromInteger(GOutbox::UNIVERSAL_MASK));
    else
        setReceivers(receivers + GPlayerSet::fromInteger(GOutbox::UNIVERSAL_MASK));
}

/** Make this a universal message (to all players). */
void
WReceiverSelector::setUniversal()
{
    setSelectedWidget();
    setReceivers(receivers + GPlayerSet::fromInteger(GOutbox::UNIVERSAL_MASK));
}

/** Set receiver set. */
void
WReceiverSelector::setReceivers(GPlayerSet set)
{
    if (receivers != set) {
        receivers = set;
        sig_receiver_changed.raise();
        drawWidget();
    }
}

void
WReceiverSelector::drawContent(GfxCanvas& can)
{
    UISkin& s = getSkin();
    GfxParameters& parms = s.getParameters();

    s.startDraw();
    parms.just_x = parms.just_y = 0;

    int y = extent.y;
    for (int i = 1; i <= 12; ++i) {
        string_t text;
        if (i == 12)
            text = _(" X - Host");
        else
            text = format(" %X - %s", i, player_racenames.getShortName(i));

        if (receivers.contains(i)) {
            parms.bgcolor = standard_colors[s.getColors()[UIColor::tc_Static]];
            parms.color   = standard_colors[s.getColors()[UIColor::tc_White]];
            parms.font    = fonts[FONT_NORMAL];
            outTextF(can, parms, getExtent().x, y, getExtent().w, text);
        } else {
            s.setSolid(true);
            s.setColor(UIColor::tc_Static);
            s.outTextF(can, FONT_NORMAL,
                       getExtent().x, y, getExtent().w, text);
        }
        y += font_heights[FONT_NORMAL];
    }

    can.update(getExtent());
}

void
WReceiverSelector::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.pref_size = li.min_size = font_sizes[FONT_NORMAL].scaledBy(17, 12);
}

bool
WReceiverSelector::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Mouse && isInside(GfxPoint(event.mouse.x, event.mouse.y))) {
        setSelectedWidget();
        if (event.mouse.b) {
            if (!mdown) {
                int i = (event.mouse.y - getExtent().y) / font_heights[FONT_NORMAL] + 1;
                if (i > 0 && i <= 12)
                    setReceivers(receivers ^ i);
            }
            mdown = true;
        } else {
            mdown = false;
        }
        return true;
    } else if (event.type == ev_Key) {
        if (event.key.code >= '1' && event.key.code <= '9') {
            setSelectedWidget();
            setReceivers(receivers ^ (event.key.code - '0'));
            return true;
        } else if (event.key.code == 'a') {
            setSelectedWidget();
            setReceivers(receivers ^ 10);
            return true;
        } else if (event.key.code == 'b') {
            setSelectedWidget();
            setReceivers(receivers ^ 11);
            return true;
        } else if (event.key.code == 'x') {
            setSelectedWidget();
            setReceivers(receivers ^ 12);
            return true;
        }
    }
    return UIWidget::handleEvent(event, second_pass);
}

/**************************** Standard Window ****************************/

namespace {
    class StandardWindow : public UIWindow {
        ObjectHolder h;
        WReceiverSelector& sel;
     public:
        StandardWindow(WReceiverSelector& win, int num_revisable);
    };
}

StandardWindow::StandardWindow(WReceiverSelector& sel, int num_revisable)
    : UIWindow(_("Send Message"), UIWindow::BLUE_WINDOW, COLOR_GRAY), sel(sel)
{
    setLayout(UIHBoxLayout::instance5);

    UIFrameGroup& fg = h.add(new UIFrameGroup(UIVBoxLayout::instance0));
    fg.add(sel);
    
    UIGroup& g1 = h.add(new UIGroup(UIVBoxLayout::instance5));
    g1.add(h.add(new UIStaticText(_("Choose receivers:"), UIStaticText::BlackTitle)));
    g1.add(fg);

    UIButton& univ = h.add(new UIButton(WReceiverSelector::sdr_Send, UIKeyString(_("Universal Message"))));
    univ.addStop();
    univ.sig_fired.add(&sel, &WReceiverSelector::setUniversal);
    g1.add(univ);
    if (num_revisable)
        g1.add(h.add(new UIButton(WReceiverSelector::sdr_Revise, UIKeyString(_("Revise messages")))).addStop());

    add(g1);

    UIGroup& g2 = h.add(new UIGroup(UIVBoxLayout::instance5));
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(WReceiverSelector::sdr_Send,   _("OK"), SDLK_RETURN)).addStop());
    g2.add(h.add(new UIButton(WReceiverSelector::sdr_Cancel, _("Cancel"), SDLK_ESCAPE)).addStop());
    add(g2);

    pack();
}

/** Standard dialog. This displays the standard 'send message' dialog
    and lets the user choose receivers.
    \param num_revisable number of existing messages. Only if some
    messages exist, this offers the 'revise' button. */
WReceiverSelector::StandardDialogResult
WReceiverSelector::doStandardDialog(int num_revisable)
{
    StandardWindow win(*this, num_revisable);
    ui_root->centerWidget(win);
    ui_root->add(win);
    StandardDialogResult rv = StandardDialogResult(ui_root->run());
    ui_root->close(win);
    if (rv == sdr_Send && getReceivers().empty())
        return sdr_Cancel;
    else
        return rv;
}

bool
WReceiverSelector::doSimpleDialog(string_t title)
{
    bool rv;
    UIWindow win(title, UIWindow::BLUE_WINDOW, COLOR_GRAY);
    UIGroup  g(UIHBoxLayout::instance5);
    UIButton uni(100, UIKeyString(_("Universal Message")));
    UIButton ok (1, _("OK"),     SDLK_RETURN);
    UIButton esc(0, _("Cancel"), SDLK_ESCAPE);
    UISpacer spc;
    ok.addStop();
    esc.addStop();
    uni.sig_fired.add(this, &WReceiverSelector::toggleUniversal);
    win.setLayout(UIVBoxLayout::instance5);
    win.add(*this);
    win.add(uni);
    g.add(ok);
    g.add(esc);
    g.add(spc);
    win.add(g);
    win.pack();
    ui_root->moveWidgetToEdge(win, 2, 2, 10);
    ui_root->add(win);
 again:
    rv = ui_root->run();
    if (rv && getReceivers().empty()) {
        messageBox(_("You must choose one or more receivers for your message."),
                   _("Edit Message"));
        goto again;
    }
    ui_root->close(win);
    return rv;
}
