/**
  *  \file pdk-host.cc
  *  \brief PDK-based FLAK server
  *
  *  This is the FLAK server based on the PDK. It's quite an ugly
  *  mixture of C and C++ - after all, the PDK is in C.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#define FLAK_USE_PDK

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <vector>
#include <cstring>
#include <algorithm>
#include <cmath>
#include <cerrno>
#include "cpluslib/assert.h"
#include "cpluslib/smallset.h"
#include "flak/flak-port.h"
#include "flak/flak.h"
#include "flak/flakconf.h"
#include "flak/flakdefs.h"
#include "game/formula.h"
#include "io/dirs.h"
#include "io/io.h"
#include "phostpdk.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "util/nls.h"

#define NEW_PDK

/* Note on temporary files: this generates a temporary file and stores
   all fights in it. It then, repeatedly, reads that file and extracts
   player information. You can therefore easily shoot yourself in your
   foot by meddling with that file while FLAK is running. I think this
   is not a major problem. One - PHost also does it, two - if you want
   to hang yourself, go ahead. This is not a security critical daemon. */

/** Convert BASIC string (from a ship/planet name field) into C
    string. C++ wizardry :-) */
#define CSTR(x) (getBasicStringN((x), 20).c_str())

static const char* program_basename;

/** Name of the temporary file on host side. */
static const char DATA_FILE_NAME[] = "flak.hst";

/** Host-side data file. This handle is used during generation of the
    combat. */
static std::FILE* data_file;

/** Verbosity level (number of -V options). */
static int verbose_level = 0;

/** Force all ships to be aggressive (-a). */
static bool aggression_flag = false;

/** Read-only flag. */
static bool readonly_flag = false;

/** Display messages on screen flag. */
static bool display_messages_flag = false;

/** Make all ships fight. */
static bool ragnarok_flag = false;

/** Total number of battles generated. */
static int battle_counter;

/** Visibility info for each fight. */
static std::vector<SmallSet<int> > battle_vis_info;

/** Display version number (-v). */
static void
version()
{
    std::cout << "FLAK - Fleet Action Kombat - Version " FLAK_VERSION "\n"
              << FLAK_COPYRIGHT << "\n"
              << "FLAK Host based on the PHost Development Kit\n"
              << "Send comments, bug reports, etc. to <" FLAK_MAIL ">\n";
    std::exit(0);
}

/** Display help screen (-h). */
static void
help()
{
    std::cout << "FLAK - Fleet Action Kombat - Version " FLAK_VERSION "\n"
                 "\n"
                 "Usage: " << program_basename << " [-hv]\n"
                 "       " << program_basename << " [-VVVZamr] [gamedir [rootdir]]\n"
                 "\n"
                 "  -h    this help screen\n"
                 "  -v    show version number\n"
                 "  -V    increase verbosity\n"
                 "  -a    force all ships to be aggressive (TESTING USE ONLY!)\n"
                 "  -m    display all messages on screen\n"
                 "  -r    read-only operation (do not modify universe)\n"
                 "  -Z    ragnarok. Make a fight involving all ships.\n"
                 "\n"
                 "The FLAK host reads its configuration from pconfig.src\n"
                 "or flak.src.\n"
                 "\n"
                 "Send comments, bug reports, etc. to <" FLAK_MAIL ">\n";
    std::exit(0);
}

/** Load configuration file. */
static void
loadFlakConfig()
{
    initFlakConfig();

    Ptr<Stream> s = Stream::openFileNT(Directory::makeFileName(gGameDirectory, "flak.src"), Stream::C_READ);
    if (s) {
        Info("Looking for configuration in `flak.src'...");
        loadFlakConfig(*s, true);
    } else {
        s = Stream::openFileNT(Directory::makeFileName(gGameDirectory, "pconfig.src"), Stream::C_READ);
        if (s) {
            Info("Looking for configuration in `pconfig.src'...");
            loadFlakConfig(*s, false);
        } else {
            ErrorExit("Unable to find configuration information.");
        }
    }

    if (verbose_level > 1) {
        Info("Configuration parameters used:");
        Info("  RatingBeamScale            = %-5d  # * beam_count * beam_type", flak_config.RatingBeamScale);
        Info("  RatingTorpScale            = %-5d  # * tube_count * torp_type", flak_config.RatingTorpScale);
        Info("  RatingBayScale             = %-5d  # * bay_count", flak_config.RatingBayScale);
        Info("  RatingMassScale            = %-5d  # * hull_mass", flak_config.RatingMassScale);
        Info("  RatingPEBonus              = %-5d  # if PE matches", flak_config.RatingPEBonus);
        Info("  RatingFullAttackBonus      = %-5d  # if no immunities", flak_config.RatingFullAttackBonus);
        Info("  RatingRandomBonus          = %-5d  # random bonus, always given", flak_config.RatingRandomBonus);
        Info("  StartingDistanceShip       = %d", flak_config.StartingDistanceShip);
        Info("  StartingDistancePlanet     = %d", flak_config.StartingDistancePlanet);
        Info("  StartingDistancePerPlayer  = %-5d  # * player_count", flak_config.StartingDistancePerPlayer);
        Info("  StartingDistancePerFleet   = %-5d  # distance between fleets", flak_config.StartingDistancePerFleet);
        Info("  CompensationShipScale      = %-5d  # * ship_count", flak_config.CompensationShipScale);
        Info("  CompensationBeamScale      = %-5d  # * beam_count", flak_config.CompensationBeamScale);
        Info("  CompensationTorpScale      = %-5d  # * tube_count", flak_config.CompensationTorpScale);
        Info("  CompensationFighterScale   = %-5d  # * bay_count", flak_config.CompensationFighterScale);
        Info("  CompensationMass100KTScale = %-5d  # * mass/100", flak_config.CompensationMass100KTScale);
        Info("  CompensationLimit          = %-5d  # maximum", flak_config.CompensationLimit);
        Info("  SendUtildata               = %s", flak_config.SendUtildata ? "yes" : "no");
        Info("  CyborgDebrisRate           = %-5d  # percent of hull minerals / fuel", flak_config.CyborgDebrisRate);
        Info("  MaximumFleetSize           = %d", flak_config.MaximumFleetSize);
        Info("Command-line options selected:");
        if (aggression_flag)
            Info("  Forced aggression (ignore alliances, force Kill mission)");
        if (ragnarok_flag)
            Info("  Ragnarok (make all ships fight)");
        if (readonly_flag)
            Info("  Read-only operation (don't modify host data, but write FLAKx.DAT files");
        if (display_messages_flag)
            Info("  Display messages on screen");
        Info("  Message level %d", verbose_level);
    }
}

/****************************** Translation ******************************/

static StringTranslator* message_translators[11];

/** Translate a message string for the given player.
    \param msg [in] The message.
    \param player [in] Player number. */
static const char*
translateMessage(const char* msg, int player)
{
    if (player > 0 && player <= 11)
        return message_translators[player-1]->translate(msg);
    else
        return msg;
}

/** Convert PHost Language_Def into two-letter language code. */
static const char*
getLanguageCodeFromPConfig(Language_Def def)
{
    switch (def) {
     case LANG_English:
     case LANG_NewEnglish:
     default:
        return "en";
     case LANG_German:
        return "de";
     case LANG_French:
        return "fr";
     case LANG_Spanish:
        return "es";
     case LANG_Italian:
        return "it";
     case LANG_Dutch:
        return "nl";
     case LANG_Russian:
        return "ru";
     case LANG_Estonian:
        return "et";
     case LANG_Polish:
        return "pl";
    }
}

/** Initialize translation system. Loads all the translation files. */
static void
initTranslation()
{
    Info("Initializing language support...");
    for (int i = 1; i <= 11; ++i) {
        try {
            message_translators[i-1] =
                loadLanguageDatabase("flak", getLanguageCodeFromPConfig(gPconfigInfo->Language[i]));
        }
        catch (...) {
            ErrorExit("Error initializing language support for player %d", i);
        }
    }
}

/***************************** Formula Stuff *****************************/

/* In FLAK, a planet fights at most once each turn. This means it
   starts its only fight with 0% damage. Therefore, all defense
   effects can be used unscaled (PHost scales down defense counts
   using the planetary damage at various places).

   We still have to handle tech levels correctly. */

/** Get damage-restricted tech level for a base. */
static int
getBaseDamageTech(int bid, BaseTech_Def level)
{
    return std::min(int((100 - BaseDamage(bid)) / 10),
                    int(BaseTech(bid, level)));
}

/** Compute number of beams on a planet. */
static int
getPlanetBeamCount(int pid)
{
    int defense = PlanetDefense(pid);
    if (IsBaseExist(pid))
        defense += BaseDefense(pid);
    defense = roundToInt(std::sqrt(defense / 3.0));
    return std::min(defense, int(CONFIG_SCALAR(AllowAlternativeCombat) ? FLAK_MAX_BEAMS : 10));
}

/** Compute beam type on a planet. */
static int
getPlanetBeamTech(int pid)
{
    int tech = roundToInt(std::sqrt(PlanetDefense(pid) / 2.0));
    if (tech > 10)
        return 10;
    else if (IsBaseExist(pid) && getBaseDamageTech(pid, BEAM_TECH) > tech)
        return getBaseDamageTech(pid, BEAM_TECH);
    else
        return tech;
}

/** Compute number of planetary tubes. */
static int
getPlanetTubeCount(int pid)
{
    if (!CONFIG_SCALAR(PlanetsHaveTubes))
        return 0;
    int defense = PlanetDefense(pid);
    if (IsBaseExist(pid))
        defense += BaseDefense(pid);
    defense = roundToInt(std::sqrt(defense / 4.0));
    return std::min(defense, int(FLAK_MAX_TORPS));
}

/** Compute torpedo type of a planet. */
static int
getPlanetTorpType(int pid)
{
    int tech = roundToInt(std::sqrt(PlanetDefense(pid) / 2.0));
    if (tech > NUM_TORPS)
        return NUM_TORPS;
    else if (IsBaseExist(pid) && getBaseDamageTech(pid, TORP_TECH) > tech)
        return getBaseDamageTech(pid, TORP_TECH);
    else
        return tech;
}

/** Compute number of torpedoes on a planet. */
static int
getPlanetTorpCount(int pid)
{
    int torps = getPlanetTubeCount(pid) * CONFIG_SCALAR(PlanetaryTorpsPerTube);
    if (IsBaseExist(pid)) {
        long sum = 0;
        for (int i = 1; i <= NUM_TORPS; ++i)
            sum += long(TorpTorpCost(i)) * BaseTorps(pid, i);
        torps += sum / TorpTorpCost(getPlanetTorpType(pid));
    }
    return torps;
}

/** Compute number of fighter bays on a planet. */
static int
getPlanetBayCount(int pid)
{
    int bays = roundToInt(std::sqrt((double) PlanetDefense(pid)));
    if (IsBaseExist(pid))
        bays += 5;
    return bays;
}

/** Compute number of fighters on a planet. */
static int
getPlanetFighterCount(int pid)
{
    int fighters = roundToInt(std::sqrt((double) PlanetDefense(pid)));
    if (IsBaseExist(pid))
        fighters += BaseFighters(pid);
    return fighters;
}

/** Compute combat mass of a planet. */
static int
getPlanetCombatMass(int pid)
{
    int mass = 100 + PlanetDefense(pid);
    if (IsBaseExist(pid))
        mass += BaseDefense(pid);
    return mass;
}


/** Get damage limit for a player (150 for lizards, 99 else). */
static inline int
getDamageLimit(int player)
{
    if (raceId(player) == 2)
        return 150;
    else
        return 99;
}


/** Scale debris amount.
    \param amount Amount of that element which was destroyed
    \return amount of that element finally reclaimed by the Borg ship. */
static int
scaleDebris(int amount)
{
    long stuff = long(amount) * flak_config.CyborgDebrisRate / 100;
    if (stuff < 0)
        return 0;
    else if (stuff > 10000)
        return 10000;
    else
        return stuff;
}

/** Predicate for std::sort(): sort an array of ship Ids so that ships
    at the same position are grouped together. */
static bool
sortByLocationPredicate(int sid1, int sid2)
{
    if (!ragnarok_flag) {
        // In Ragnarok mode, we pretend everyone's at the same place,
        // so ignore the positions.
        int diff;
        if ((diff = ShipLocationX(sid1) - ShipLocationX(sid2)) != 0)
            return diff < 0;
        if ((diff = ShipLocationY(sid1) - ShipLocationY(sid2)) != 0)
            return diff < 0;
    }
    return sid1 < sid2;
}

/***************************** Initialisation ****************************/

namespace {
    struct ShipInfo {
        /** FCBO value, plus 100. */
        int fcbo_plus_100;
        /** Ship data. */
        TFlakShip data;

        ShipInfo& initFromShip(int sid);
        ShipInfo& initFromPlanet(int pid);

        /** True iff this is a planet. */
        bool  isPlanet() const
            { return data.flags & flak_IsPlanet; }
     private:
//         void  initCommon();
//         int32 computeRating() const;
//         int32 computeStrength() const;
        int   computeFCBO(const char* fc) const;
    };
}

/** Initialize data for a ship. */
ShipInfo&
ShipInfo::initFromShip(int sid)
{
    const int level = ExperienceLevel(ShipExperience(sid));

    storeBasicString(data.name, ShipName(sid, 0));
    data.damage_init   = ShipDamage(sid);
    data.crew_init     = ShipCrew(sid);
    data.id            = sid;
    data.player        = ShipOwner(sid);
    data.hull          = ShipHull(sid);
    data.level         = level;
    data.beam_count    = ShipBeamNumber(sid);
    data.beam_type     = ShipBeamType(sid);
    data.torp_lcount   = ShipTubeNumber(sid);
    data.torp_count    = ShipTubeNumber(sid) ? ShipAmmunition(sid) : 0;
    data.torp_type     = ShipTorpType(sid);
    data.bay_count     = ShipBays(sid);
    data.fighter_count = ShipTubeNumber(sid) ? 0 : ShipAmmunition(sid);
    data.mass          = HullMass(ShipHull(sid));
    data.shield_init   = ShipCombatShieldLevel(sid);

    /* NTP */
    if (std::strcmp(ShipFCode(sid, 0), "NTP") == 0)
        data.fighter_count = data.torp_count = 0;

    /* ESB */
    int esb = 0;
    if (CONFIG_SCALAR(AllowEngineShieldBonus))
        esb += CONFIG_SCALAR(EngineShieldBonusRate);
    if (level)
        esb += gPconfigInfo->EModEngineShieldBonusRate[level-1];
    if (esb)
        data.mass += EngMoneyCost(ShipEngine(sid)) * esb / 100;

    /* Fed crew bonus */
    if (CONFIG_SCALAR(AllowFedCombatBonus) && raceId(data.player) == 1)
        data.mass += 50;

    /* extra bays */
    if (data.bay_count) {
        data.bay_count += CONFIG_ARRAY(ExtraFighterBays, data.player);
        if (level)
            data.bay_count += gPconfigInfo->EModExtraFighterBays[level-1];
        if (data.bay_count > FLAK_MAX_BAYS)
            data.bay_count = FLAK_MAX_BAYS;
    }

    data.flags = 0;  // not a planet
    fcbo_plus_100 = computeFCBO(ShipFCode(sid, 0));

    FlakBattle::initShip(data);

    return *this;
}

/** Initialize data from planet. */
ShipInfo&
ShipInfo::initFromPlanet(int pid)
{
    const int level = ExperienceLevel(PlanetExperience(pid));

    storeBasicString(data.name, PlanetName(pid, 0));
    data.damage_init   = 0; /* planet starts with 0 damage in every turn */
    data.crew_init     = 0;
    data.id            = pid;
    data.player        = PlanetOwner(pid);
    data.hull          = 0;
    data.level         = level;
    data.beam_count    = getPlanetBeamCount(pid);
    data.beam_type     = getPlanetBeamTech(pid);
    data.torp_lcount   = getPlanetTubeCount(pid);
    data.torp_count    = getPlanetTorpCount(pid);
    data.torp_type     = getPlanetTorpType(pid);
    data.bay_count     = getPlanetBayCount(pid);
    data.fighter_count = getPlanetFighterCount(pid);
    data.mass          = getPlanetCombatMass(pid);
    data.shield_init   = 100; /* planet starts with 100 shield in every turn */

    /* extra bays */
    if (data.bay_count) {
        if (level)
            data.bay_count += gPconfigInfo->EModExtraFighterBays[level-1];
        if (data.bay_count > FLAK_MAX_BAYS)
            data.bay_count = FLAK_MAX_BAYS;
    }

    data.flags = flak_IsPlanet;
    fcbo_plus_100 = computeFCBO(PlanetFCode(pid, 0));
    FlakBattle::initShip(data);

    return *this;
}

// /** Common part of initialisation. Called from
//     initFromPlanet()/initFromShip(). */
// void
// ShipInfo::initCommon()
// {
//     data.max_fighters_launched = 2 * data.bay_count;
//     data.max_fighters_launched = FLAK_MFL_SCALE * data.bay_count;
//     if (data.max_fighters_launched < FLAK_MINIMUM_MFL)
//         data.max_fighters_launched = FLAK_MINIMUM_MFL;
//     if (data.max_fighters_launched > FLAK_MAXIMUM_MFL)
//         data.max_fighters_launched = FLAK_MAXIMUM_MFL;
//     data.rating              = computeRating();
//     data.compensation_rating = computeStrength();
//     data.ending_status       = 0;
// }

/** Compute FCBO value. This returns the numerical FC value
    \em{plus 100, trimmed at 1099}. */
int
ShipInfo::computeFCBO(const char* fc) const
{
    if (((fc[0] >= '0' && fc[0] <= '9') || fc[0] == '-')
        && fc[1] >= '0' && fc[1] <= '9' && fc[2] >= '0' && fc[2] <= '9')
    {
        /* numeric fc */
        const int cent = 10 * (fc[1] - '0') + fc[2] - '0';
        if (fc[0] == '-')
            return 100 - cent;
        else
            return 100 * (fc[0] - '0') + cent + 100;
    } else {
        if ((data.flags & flak_IsPlanet) != 0
            && (std::strcmp(fc, "ATT") == 0 || std::strcmp(fc, "NUK") == 0))
            return 100;
        else
            return 1099;
    }
}

// /** Compute targeting rating. */
// int32
// ShipInfo::computeRating() const
// {
//     return data.mass * flak_config.RatingMassScale
//         + data.torp_lcount * data.torp_type * flak_config.RatingTorpScale
//         + data.beam_count * data.beam_type * flak_config.RatingBeamScale
//         + data.bay_count * flak_config.RatingBayScale;
// }

// /** Compute compensation rating. */
// int32
// ShipInfo::computeStrength() const
// {
//     int32 strength = flak_config.CompensationShipScale
//         + data.torp_lcount * flak_config.CompensationTorpScale
//         + data.beam_count * flak_config.CompensationBeamScale
//         + data.bay_count * flak_config.CompensationFighterScale
//         + data.mass * flak_config.CompensationMass100KTScale / 100;
//     return (strength > flak_config.CompensationLimit
//             ? flak_config.CompensationLimit
//             : strength < 0
//             ? 0
//             : strength);
// }

namespace {
    /** Predicate for sort: sort a list of ShipInfo into their combat
        order. Groups players together, randomly permuted. Upon
        construction, this computes a permutation and uses it later
        on. */
    class PermutedSorter {
        int player_map[NUM_OWNERS];
        int frob(int pid)
            {
                if (pid > 0 && pid <= NUM_OWNERS)
                    return player_map[pid-1];
                return pid;
            }
     public:
        PermutedSorter();
        bool operator()(const ShipInfo& l, const ShipInfo& r)
            {
                int dif = frob(l.data.player) - frob(r.data.player);
                if (dif)
                    return dif < 0;
                dif = l.fcbo_plus_100 - r.fcbo_plus_100;
                if (dif)
                    return dif < 0;
                dif = l.data.id - r.data.id;
                if (dif)
                    return dif < 0;
                return (r.data.flags & flak_IsPlanet) == 0;
            }
    };
}

/** Constructor. Generates a permutation. */
PermutedSorter::PermutedSorter()
{
    for (int i = 1; i <= NUM_OWNERS; ++i)
        player_map[i-1] = i;
    for (int i = 1; i < NUM_OWNERS; ++i) {
        int j = RandomRange(i+1);
        std::swap(player_map[i], player_map[j]);
    }
}

/** Check whether /me/ can attack /them/. This relation is NOT
    symmetric, it only handles the case where /me/ is the
    aggressor. */
static bool
isAttacking(const FlakShip& me, const FlakShip& them)
{
//     Info("isAttacking(%s, %s)", CSTR(me.data.name), CSTR(them.data.name));
//     struct foo {
//         bool ok;
//         foo() : ok(false) { }
//         ~foo() { if (!ok) Info(" -> NOT"); else Info(" -> YES"); }
//     } bar;

    char me_fc[4], them_fc[4];
    me.isPlanet()   ? PlanetFCode(me.data.id,   me_fc)   : ShipFCode(me.data.id,   me_fc);
    them.isPlanet() ? PlanetFCode(them.data.id, them_fc) : ShipFCode(them.data.id, them_fc);

    /* same race? */
    if (me.data.player == them.data.player)
        return false;

    /* if we are not aggressive, we can not attack */
    if (me.isPlanet()) {
        /* planet needs ATT/NUK to attack */
        if (!aggression_flag) {
            if (std::strcmp(me_fc, "ATT") != 0 && std::strcmp(me_fc, "NUK") != 0)
                return false;
        }
        /* planet cannot attack if so configured */
        if (!CONFIG_SCALAR(AllowPlanetAttacks))
            return false;
        /* planet cannot attack SSD/Klingon/Rebel */
        /* FIXME: update this for new hullfuncs */
        if (ShipDoesImperialAssault(them.data.id))
            return false;
        if (raceId(them.data.player) == 4 && !CONFIG_SCALAR(PlanetsAttackKlingons))
            return false;
        if (raceId(them.data.player) == 10 && !CONFIG_SCALAR(PlanetsAttackRebels))
            return false;
    } else {
        /* ship needs Kill or matching PE to attack */
        if (! aggression_flag) {
#ifdef NEW_PDK
            if (!ShipHasEnemy(me.data.id, RaceType_Def(them.data.player)))
                return false;
#else
            if (ShipMission(me.data.id) != Kill) {
                if (ShipEnemy(me.data.id) != them.data.player)
                    return false;
            }
#endif
        }
        /* cloaked ships can not attack planets (or not attack at all) */
        if (IsShipCloaked(me.data.id) && (!CONFIG_SCALAR(AllowCloakedShipsAttack) || them.isPlanet()))
            return false;
        /* fuelless ships cannot attack */
        if (ShipCargo(me.data.id, NEUTRONIUM) == 0)
            return false;
    }

    /* you cannot attack cloaked ships. In PHost, we *can* attack a
       cloaked ship if it has us as its PE. This mainly has an effect
       on battle order, so we do not emulate that goof. */
    if (!them.isPlanet() && IsShipCloaked(them.data.id))
        return false;

    /* allied? */
    if (! aggression_flag) {
        if (PlayersAreAllies(RaceType_Def(me.data.player), RaceType_Def(them.data.player))
            && PlayerAllowsAlly(RaceType_Def(me.data.player), RaceType_Def(them.data.player), ALLY_COMBAT))
            return false;
    }

    /* do not attack with ineffective weapons. NEW: we also do not
       attack when we don't have ammo. */
    // FIXME: this test also appears inside the player (canStillAttack)
    if (them.isPlanet()) {
        if (me.data.bay_count == 0 || me.data.fighter_count == 0) {
            int tc = me.data.torp_lcount, bc = me.data.beam_count;
            if (tc && (getTorp(me.data.torp_type).getDamage() == 0 || me.data.torp_count == 0))
                tc = 0;
            if (bc && getBeam(me.data.beam_type).getDamage() == 0)
                bc = 0;
            if (tc == 0 && bc == 0)
                return false;
        }
    } else {
        if (me.data.beam_count == 0
            && (me.data.bay_count == 0 || me.data.fighter_count == 0)
            && (me.data.torp_lcount == 0 || me.data.torp_count == 0))
            return false;
    }

    /* we cannot attack fuelless ships, except when we are a NUKing
       planet, and they are not Bird capital ships */
    if (!them.isPlanet() && ShipCargo(them.data.id, NEUTRONIUM) == 0) {
        if (!me.isPlanet() || std::strcmp(me_fc, "NUK") != 0)
            return false;
        if (raceId(them.data.player) == 3 && (them.data.beam_count || them.data.bay_count || them.data.torp_lcount))
            return false;
    }

    /* matching fcode? */
    if (std::strcmp(me_fc, them_fc) == 0 && !IsFCodeSpecial(me_fc))
        return false;

//     bar.ok = true;
    return true;
}

/** Check whether /ship/ is unarmed. */
static inline bool
isUnarmed(const FlakShip& ship)
{
    return ship.data.beam_count == 0 && ship.data.torp_lcount == 0 && ship.data.bay_count == 0;
}

/** Check whether any ship from me can attack any ship from them. */
static bool
canAttackThisFleet(const FlakBattle& battle, const FlakFleet& me, const FlakFleet& them)
{
    // shortcut
    if (&me == &them)
        return false;

    // check it
    for (int my_index = 0; my_index < me.data.num_ships; ++my_index) {
        for (int their_index = 0; their_index < them.data.num_ships; ++their_index) {
            const FlakShip& my_ship = battle.getShipByNumber(me.data.first_ship + my_index);
            const FlakShip& their_ship = battle.getShipByNumber(them.data.first_ship + their_index);
            if (isAttacking(my_ship, their_ship) || isAttacking(their_ship, my_ship))
                if (!(isUnarmed(my_ship) && isUnarmed(their_ship)))
                    return true;
        }
    }
    return false;
}

/** Compute attack list for one fleet. */
static void
computeAttackList(FlakBattle& battle, FlakFleet& fleet)
{
    int32 first_entry = battle.getAttackListNumber();

#if FLAK_ALGORITHM >= 20060531
    /* Fleet/fleet attack relations must be symmetrical. If any ship
       from a fleet can attack/be attacked by us, we must be allowed
       to attack all ships from that fleet. Otherwise, it could happen
       that the other fleet is still aggressive to us, but we do no
       longer have someone to attack. */
    for (int other_fleet_nr = 0; other_fleet_nr < battle.getNumFleets(); ++other_fleet_nr) {
        const FlakFleet& other_fleet = battle.getFleetByNumber(other_fleet_nr);
        if (canAttackThisFleet(battle, fleet, other_fleet)) {
            for (int other_index = 0; other_index < other_fleet.data.num_ships; ++other_index) {
                const int other_member = other_index + other_fleet.data.first_ship;
                const FlakShip& them = battle.getShipByNumber(other_member);
                bool can_attack = false, match_pe = false, match_fc = false;
                // for each fleet member
                for (int fleet_index = 0; fleet_index < fleet.data.num_ships; ++fleet_index) {
                    const int fleet_member = fleet_index + fleet.data.first_ship;
                    const FlakShip& me = battle.getShipByNumber(fleet_member);
                    if (isAttacking(me, them) || isAttacking(them, me)) {
                        if (!(isUnarmed(me) && isUnarmed(them))) {
                            // note that we deliberately do not handle
                            // permanent primary enemies ('enemies' command)
                            // here.
                            can_attack = true;
                            if (!me.isPlanet() && ShipEnemy(me.data.id) == them.data.player)
                                match_pe = true;
                        }
                    } else {
                        match_fc = true;
                    }
                }
                if (can_attack) {
                    // we can attack it regularily
                    int bonus = RandomRange(flak_config.RatingRandomBonus);
                    if (match_pe)
                        bonus += flak_config.RatingPEBonus;
                    if (!match_fc)
                        bonus += flak_config.RatingFullAttackBonus;
                    if (!bonus)
                        bonus = 1;
                    battle.addAttackListEntry(other_member, bonus);
                } else {
                    // we cannot attack it, so give it priority 0
                    battle.addAttackListEntry(other_member, 0);
                }
            }
        }
    }
#else
    // for each ship (=possible opponent of fleet)
    for (int possible_opponent = 0; possible_opponent < battle.getNumShips(); ++possible_opponent) {
        bool can_attack = false, match_pe = false, match_fc = false;
        // for each fleet member
        for (int fleet_index = 0; fleet_index < fleet.data.num_ships; ++fleet_index) {
            const int fleet_member = fleet_index + fleet.data.first_ship;
            const FlakShip& me = battle.getShipByNumber(fleet_member);
            const FlakShip& them = battle.getShipByNumber(possible_opponent);
            if (isAttacking(me, them) || isAttacking(them, me)) {
                if (!(isUnarmed(me) && isUnarmed(them))) {
                    // note that we deliberately do not handle
                    // permanent primary enemies ('enemies' command)
                    // here.
                    can_attack = true;
                    if (!me.isPlanet() && ShipEnemy(me.data.id) == them.data.player)
                        match_pe = true;
                }
            } else {
                match_fc = true;
            }
        }
        if (can_attack) {
            int bonus = RandomRange(flak_config.RatingRandomBonus);
            if (match_pe)
                bonus += flak_config.RatingPEBonus;
            if (!match_fc)
                bonus += flak_config.RatingFullAttackBonus;
            battle.addAttackListEntry(possible_opponent, bonus);
        }
    }
#endif
    fleet.data.att_list_pointer = first_entry;
    fleet.data.att_list_size    = battle.getAttackListNumber() - first_entry;
}

/******************************** Messaging ******************************/

/** Send message to a player. */
static void
doMessage(int player, string_t text)
{
    /* replace \r by \n */
    for (string_t::size_type i = text.find('\n'); i != text.npos; i = text.find('\n', i+1))
        text[i] = '\r';

    /* write it out */
    if (!readonly_flag)
        WriteAUXHOSTMessage(RaceType_Def(player), text.c_str());
    if (display_messages_flag) {
        while (text.size() && text[text.size()-1] == '\r')
            text.erase(text.size()-1);
        for (string_t::size_type i = text.size(); i > 0; --i)
            if (text[i-1] == '\r')
                text.replace(i-1, 1, "\nSubspace Message: ");
        Info("Subspace Message: %s", text.c_str());
    }
}

/** Format a battle result message. The format string can contain
    escape sequences:
    - %n1    name and Id of object 1
    - %i1    Id of object 1
    - %t1    type ('ship', 'planet') of object 1
    - %o1    owner of object 1
    - %h1    Message header letter + Id of object 1
    - same ending with 2 for object 2.

    \param fmt     format string (untranslated)
    \param one,two participants
    \param player  receiver of message */
static string_t
formatMessage(const char* fmt, const FlakShip& one, const FlakShip& two, int player)
{
    string_t result;
    fmt = translateMessage(fmt, player);
    while (const char* p = std::strchr(fmt, '%')) {
        ASSERT(p[1]);
        ASSERT(p[2] == '1' || p[2] == '2');
        const FlakShip& o = (p[2] == '1' ? one : two);

        result.append(fmt, p - fmt);
        fmt = p+3;
        switch (p[1]) {
         case 'n':
            result.append(getBasicString(o.data.name));
            result.append(format(" (#%d)", o.data.id));
            break;
         case 'i':
            result.append(format("%d", o.data.id));
            break;
         case 't':
            if (o.isPlanet())
                result.append(translateMessage("planet", player));
            else
                result.append(translateMessage("ship", player));
            break;
         case 'o':
            result.append(RaceNameAdjective(RaceType_Def(o.data.player), 0));
            break;
         case 'h':
            if (o.isPlanet())
                result.append(format("(-p%04d)", o.data.id));
            else
                result.append(format("(-s%04d)", o.data.id));
            break;
         default:
            ASSERT(0);
        }
    }
    result.append(fmt);
    return result;
}

/******************************** Results ********************************/

namespace {
    /** Result for once race. */
    struct RaceResult {
        int  num_ships;             ///< Total number of units this race starts with.
        long total_experience;      ///< Total experience gathered by this race in this fight.
        long total_pbps_times_1000; ///< Total PBPs gathered by this race, times 1000.
        long total_exp_mass;        ///< Sum of this race's masses, for experience calculation.
        RaceResult()
            : num_ships(0), total_experience(0), total_pbps_times_1000(0),
              total_exp_mass(0)
            { }
    };
}

/** Ship gather debris.
    \param victor a surviving cyborg ship
    \param wreck  the destroyed enemy */
static void
doGatherDebris(const FlakShip& victor, const FlakShip& wreck)
{
    const int vsid = victor.data.id, wsid = wreck.data.id, whull = ShipHull(wsid);

    int free = HullFuelCapacity(ShipHull(vsid)) - ShipCargo(vsid, NEUTRONIUM);
    int have = scaleDebris(ShipCargo(wsid, NEUTRONIUM));
    have = std::min(have, free);
    PutShipCargo(vsid, NEUTRONIUM, ShipCargo(vsid, NEUTRONIUM) + have);

    free = HullCargoCapacity(ShipHull(vsid)) - ShipCargoMass(vsid);
    have = std::min(free, scaleDebris(HullTritCost(whull)));
    PutShipCargo(vsid, TRITANIUM, ShipCargo(vsid, TRITANIUM) + have);
    free -= have;

    have = std::min(free, scaleDebris(HullDurCost(whull)));
    PutShipCargo(vsid, DURANIUM, ShipCargo(vsid, DURANIUM) + have);
    free -= have;

    have = std::min(free, scaleDebris(HullMolyCost(whull)));
    PutShipCargo(vsid, MOLYBDENUM, ShipCargo(vsid, MOLYBDENUM) + have);
    free -= have;

    doMessage(victor.data.player,
              formatMessage("%h1<<< Ship Message >>>\n\n"
                            "FROM: %n1\n\n"
                            "Beaming enemy ship debris\n"
                            "into cargo hold.\n"
                            "Gathering Neutronium.\n",
                            victor, wreck,
                            victor.data.player));
}

/** Handle common parts of planet capture.
    - destroy base
    - natives get unhappy
    - fcode randomize */
static void
handlePlanetCapture(int pid)
{
    if (IsBasePresent(pid))
        DeleteBase(pid);

    if (PlanetHasNatives(pid)) {
        PutPlanetNatHappy(pid, PlanetNatHappy(pid) - 20);
        PutPlanetNativePopulation(pid, PlanetNativePopulation(pid) * long(gPconfigInfo->NativeCombatSurvivalRate) / 100);
        if (PlanetNativePopulation(pid) < 100) {
            PutPlanetNatTax(pid, 0);
            PutPlanetNatHappy(pid, 100);
            PutPlanetNativePopulation(pid, 0);
            PutPlanetNativeType(pid, NoNatives);
            PutPlanetNatGovm(pid, NoGovm);
        }
    }

    char buf[4];
    sprintf(buf, "%03u", RandomRange(1000));
    PutPlanetFCode(pid, buf);
}

/** Check whether points should be given for a fight between a and b. 
    No points are given if players are allied, but full points are
    given if both are owned by same player (capture back). */
static bool
givePointsForFight(const FlakShip& a, const FlakShip& b)
{
    return (a.data.player == b.data.player)
        || !PlayersAreAllies(RaceType_Def(a.data.player), RaceType_Def(b.data.player));
}

/** Process result for one planet.
    \pre pl.isPlanet()
    \pre results points to 11 RaceResult's. */
static void
processPlanetResult(FlakBattle& b, FlakShip& pl, RaceResult results[])
{
    const int pid = pl.data.id;

    /* count it */
    results[pl.data.player-1].total_exp_mass += pl.data.mass - 100;
    results[pl.data.player-1].num_ships ++;

    /* Handle damage to planet */
    int p_damage = int(pl.status.damage);
    if (p_damage > 100)
        p_damage = 100;
    PutPlanetDefense(pid, PlanetDefense(pid) * long(100 - pl.status.damage) / 100);

    if (IsBaseExist(pid)) {
        /* Handle ammunition differences. */
        int fighters = BaseFighters(pid) + pl.data.fighter_count - pl.status.fighter_count;
        if (fighters < 0)
            PutBaseFighters(pid, 0);
        else if (fighters > MAX_BASE_FIGHTERS)
            PutBaseFighters(pid, MAX_BASE_FIGHTERS);
        else
            PutBaseFighters(pid, fighters);

        if (CONFIG_SCALAR(PlanetsHaveTubes)) {
            int32 torp_cost = int32(pl.data.torp_count - pl.status.torp_count) * TorpTorpCost(pl.data.torp_type);
            bool  did_some;
            while (torp_cost > 0) {
                did_some = false;
                for (int t = 1; t <= NUM_TORPS; ++t) {
                    if (BaseTorps(pid, t) > 0 && TorpTorpCost(t) <= torp_cost) {
                        PutBaseTorps(pid, t, BaseTorps(pid, t) - 1);
                        torp_cost -= TorpTorpCost(t);
                        did_some = true;
                    }
                }
                if (! did_some)
                    break;
            }
        }

        /* Add base damage */
        int new_damage = BaseDamage(pid) + p_damage;
        if (new_damage >= 100) {
            DeleteBase(pid);
        } else {
            PutBaseDamage(pid, new_damage);
            PutBaseDefense(pid, BaseDefense(pid) * (100 - p_damage) / 100);

            int max_tech = (100 - new_damage) / 10;
            for (int i = 0; i < 4; ++i)
                if (BaseTech(pid, BaseTech_Def(i)) > max_tech)
                    PutBaseTech(pid, BaseTech_Def(i), max_tech);
        }
    }

    /* Now process result. */
    if (pl.isAlive()) {
        if (verbose_level > 2)
            Info("      Planet #%d '%s' survived", pl.data.id, CSTR(pl.data.name));
        pl.data.ending_status = 0;
    } else if (const FlakShip* opp_ptr = b.findCaptor(pl)) {
        /* planet got captured. */
        const FlakShip& opp = *opp_ptr;

        if (verbose_level > 2)
            Info("      Planet #%d '%s' captured by #%d '%s'",
                 pl.data.id, CSTR(pl.data.name), opp.data.id, CSTR(opp.data.name));

        /* tons */
        if (gPconfigInfo->TonsScoreCountsPlanets && pl.data.player != opp.data.player) {
#ifdef NEW_PDK
            AddToTonsScore(RaceType_Def(opp.data.player),  // winner
                           RaceType_Def(pl.data.player),   // loser
                           pl.data.mass);
            // results[pl.data.player-1].tons_lost += pl.data.mass;
            // results[opp.data.player-1].tons_destroyed += pl.data.mass;
#endif
        }

        /* opponent receives PBPs and exp. We do give exp/PBP for capturing back. */
        if (givePointsForFight(pl, opp)) {
            int32 mass = pl.data.mass - 100;
            int scale = IsBaseExist(pid)
                ? CONFIG_ARRAY(PALCombatBaseScaling, opp.data.player)
                : CONFIG_ARRAY(PALCombatPlanetScaling, opp.data.player);
            results[opp.data.player-1].total_pbps_times_1000 +=
                scale * mass * (CONFIG_ARRAY(PALAggressorPointsPer10KT, opp.data.player)
                                + CONFIG_ARRAY(PALAggressorKillPointsPer10KT, opp.data.player));
            results[opp.data.player-1].total_experience +=
                mass * (CONFIG_SCALAR(EPCombatDamageScaling) + CONFIG_SCALAR(EPCombatKillScaling));
        }

        /* now do the trading */
        ChangePlanetOwner(pid, RaceType_Def(opp.data.player));
        PutPlanetDefense(pid, 0);
        pl.data.ending_status = opp.data.player;

        uint32 colos = gPconfigInfo->ColonistCombatSurvivalRate * PlanetCargo(pid, COLONISTS) / 100;
        if (colos < 100)
            colos = 100;
        PutPlanetCargo(pid, COLONISTS, colos);
        PutPlanetColHappy(pid, PlanetColHappy(pid) - gPconfigInfo->ColonistCombatSurvivalRate / 2);

        Exp_Handle pexp = PlanetExperience(pid);
        PutExperiencePoints(pexp, ExperiencePoints(pexp) * gPconfigInfo->ColonistCombatSurvivalRate / 100);

        handlePlanetCapture(pid);

        if (opp.data.player == pl.data.player) {
            doMessage(pl.data.player,
                      formatMessage("%h1<<< Fleet Message >>>\n\n"
                                    "The %t1 %n1\n"
                                    "got captured in a battle, but\n"
                                    "%t2 %n2\n"
                                    "captured it back\n",
                                    pl, opp,
                                    pl.data.player));
        } else {
            doMessage(pl.data.player,
                      formatMessage("%h1<<< Fleet Message >>>\n\n"
                                    "The %t1 %n1\n"
                                    "got captured by\n"
                                    "the %o2 %t2\n"
                                    "%n2\n",
                                    pl, opp,
                                    pl.data.player));
            doMessage(opp.data.player,
                      formatMessage("%h1<<< Fleet Message >>>\n\n"
                                    "The %t2 %n2\n"
                                    "has captured\n"
                                    "the %o1 %t1\n"
                                    "%n1\n",
                                    pl, opp,
                                    opp.data.player));
        }
    } else {
        if (verbose_level > 2)
            Info("      Planet #%d '%s' unowned", pl.data.id, CSTR(pl.data.name));

        /* planet is left and everyone's dead. Make it unowned */
        ChangePlanetOwner(pid, NoRace);
        PutPlanetCargo(pid, COLONISTS, 0);
        PutPlanetColTax(pid, 0);
        PutPlanetColHappy(pid, 0);
        PutPlanetDefense(pid, 0);

        handlePlanetCapture(pid);

        doMessage(pl.data.player,
                  formatMessage("%h1<<< Fleet Message >>>\n\n"
                                "The %n1 %t1\n"
                                "got destroyed in battle\n",
                                pl, pl,
                                pl.data.player));

        pl.data.ending_status = -1;
    }
}

/** Process result for one ship. */
static void
processShipResult(FlakBattle& b, FlakShip& sh, RaceResult results[])
{
    const int sid = sh.data.id;

    /* count it */
    const int32 mass = gPconfigInfo->PALIncludesESB[sh.data.player] ? sh.data.mass : HullMass(ShipHull(sid));
    results[sh.data.player-1].total_exp_mass += mass;
    results[sh.data.player-1].num_ships++;

    /* handle ammunition difference */
    int ammo_lost = 0;
    if (sh.data.bay_count > 0)
        ammo_lost = sh.data.fighter_count - sh.status.fighter_count;
    else if (sh.data.torp_lcount > 0)
        ammo_lost = sh.data.torp_count - sh.status.torp_count;
    if (ammo_lost > 0) {
        int new_ammo = ShipAmmunition(sid) - ammo_lost;
        if (new_ammo < 0)
            PutShipAmmunition(sid, 0);
        else
            PutShipAmmunition(sid, new_ammo);
    } else if (ammo_lost < 0) {
        // be careful not to overload a ship. It might have got new
        // ammo if it is a carrier (since 20060531 also if it is a torper)
        int free_cargo = HullCargoCapacity(ShipHull(sid)) - ShipCargoMass(sid);
        if (free_cargo > 0) {
            if (-ammo_lost > free_cargo)
                ammo_lost = -free_cargo;
            PutShipAmmunition(sid, ShipAmmunition(sid) - ammo_lost);
        }
    }

    /* other specs */
    PutShipCrew(sid, std::max(int(sh.status.crew + 0.5), 0));
    PutShipDamage(sid, std::min(int(sh.status.damage + 0.5), 150));
    SetShipCombatShieldLevel(sid, int(sh.status.shield));

    const FlakShip* opp_ptr;
    if (sh.isAlive()) {
        /* ship still alive. Nothing to do. */
        if (verbose_level > 2)
            Info("      Ship #%d '%s' survived", sh.data.id, CSTR(sh.data.name));
        sh.data.ending_status = 0;
    } else if ((opp_ptr = b.findCaptor(sh)) != 0
               && ShipCrew(sid) == 0
               && ShipDamage(sid) < getDamageLimit(sh.data.player)
               && ShipDamage(sid) < getDamageLimit(opp_ptr->data.player))
    {
        /* captured */
        const FlakShip& opp = *opp_ptr;

        if (verbose_level > 2)
            Info("      Ship #%d '%s' captured by #%d '%s'",
                 sh.data.id, CSTR(sh.data.name), opp.data.id, CSTR(opp.data.name));

        PutShipOwner(sid, RaceType_Def(opp.data.player));
        sh.data.ending_status = opp.data.player;
        int new_crew = HullCrewComplement(ShipHull(sid));
        if (new_crew > 10)
            new_crew = 10;
        PutShipCrew(sid, new_crew);
        if (ShipMission(sid) == Kill)
            PutShipMission(sid, NoMission);
        PutShipEnemy(sid, NoRace);

        /* experience for recrewing */
        int32 opp_exp = ExperiencePoints(opp.isPlanet()
                                         ? PlanetExperience(opp.data.id)
                                         : ShipExperience(opp.data.id));
        int rate = (opp.data.player > 0 && opp.data.player <= NUM_PLAYERS)
            ? CONFIG_ARRAY(EPRecrewScaling, opp.data.player) : 0;
        PutExperiencePoints(ShipExperience(sid),
                            (opp_exp * (100-rate) + ExperiencePoints(ShipExperience(sid)) * rate) / 100);

        /* bonuses */
        if (givePointsForFight(sh, opp)) {
            /* bonuses */
            int damage_done = ShipDamage(sid) - sh.data.damage_init;
            results[opp.data.player-1].total_pbps_times_1000 +=
                mass * CONFIG_ARRAY(PALAggressorPointsPer10KT, opp.data.player) * damage_done
                + 100L * sh.data.crew_init * CONFIG_ARRAY(PALShipCapturePer10Crew, opp.data.player);
            results[opp.data.player-1].total_experience +=
                mass * CONFIG_SCALAR(EPCombatDamageScaling) * damage_done / 100;
        }

        /* messages */
        if (opp.data.player == sh.data.player) {
            doMessage(sh.data.player,
                      formatMessage("%h1<<< Fleet Message >>>\n\n"
                                    "The %t1 %n1\n"
                                    "got captured in battle, but\n"
                                    "%t2 %n2\n"
                                    "captured it back\n",
                                    sh, opp,
                                    sh.data.player));
        } else {
            doMessage(sh.data.player,
                      formatMessage("%h1<<< Fleet Message >>>\n\n"
                                    "The %t1 %n1\n"
                                    "got captured by\n"
                                    "the %o2 %t2\n"
                                    "%n2\n",
                                    sh, opp,
                                    sh.data.player));
            doMessage(opp.data.player,
                      formatMessage("%h1<<< Fleet Message >>>\n\n"
                                    "The %t2 %n2\n"
                                    "has captured\n"
                                    "the %o1 %t1\n"
                                    "%n1\n",
                                    sh, opp,
                                    opp.data.player));
        }
    } else {
        /* destroyed. */
        const FlakShip& opp = *sh.status.last_hit_by;

        if (verbose_level > 2)
            Info("      Ship #%d '%s' destroyed by #%d '%s'",
                 sh.data.id, CSTR(sh.data.name), opp.data.id, CSTR(opp.data.name));

        if (givePointsForFight(sh, opp)) {
            const int damage_done = ShipDamage(sid) - sh.data.damage_init;
            results[opp.data.player-1].total_pbps_times_1000 +=
                mass * CONFIG_ARRAY(PALAggressorPointsPer10KT, opp.data.player) * damage_done
                + mass * CONFIG_ARRAY(PALAggressorKillPointsPer10KT, opp.data.player) * 100;
            results[opp.data.player-1].total_experience +=
                mass * CONFIG_SCALAR(EPCombatDamageScaling) * damage_done / 100
                + mass * CONFIG_SCALAR(EPCombatKillScaling);
#ifdef NEW_PDK
            AddToTonsScore(RaceType_Def(opp.data.player),    // winner
                           RaceType_Def(sh.data.player),     // loser
                           mass);
            // results[opp.data.player-1].tons_destroyed += mass;
            // results[sh.data.player-1].tons_lost += mass;
#endif
        }

        sh.data.ending_status = -1;

        doMessage(sh.data.player,
                  formatMessage("%h1<<< Fleet Message >>>\n\n"
                                "Our %t1 %n1\n"
                                "was destroyed by\n"
                                "the %o2 %t2\n"
                                "%n2\n",
                                sh, opp,
                                sh.data.player));
        doMessage(opp.data.player,
                  formatMessage("%h2<<< Fleet Message >>>\n\n"
                                "The %t2 %n2\n"
                                "has destroyed\n"
                                "the %o1 %t1\n"
                                "%n1\n",
                                sh, opp,
                                opp.data.player));

        /* debris */
        if (opp.isAlive() && !opp.isPlanet() && raceId(opp.data.player) == 6)
            doGatherDebris(opp, sh);

        DeleteShip(sid);
    }
}

/** Given a completed battle, incorporate result back into game data. */
static void
resolveBattle(FlakBattle& battle)
{
    RaceResult results[NUM_OWNERS];

    for (int i = 0; i < battle.getNumShips(); ++i) {
        FlakShip& sh = battle.getShipByNumber(i);
        if (sh.isPlanet())
            processPlanetResult(battle, sh, results);
        else
            processShipResult(battle, sh, results);
    }

    /* give bonuses */
    if (verbose_level > 2)
        Info("    Computing scores...");
    for (int i = 0; i < battle.getNumShips(); ++i) {
        FlakShip& sh = battle.getShipByNumber(i);
        if (sh.isAlive()) {
            if (results[sh.data.player-1].total_exp_mass) {
                int new_exp = results[sh.data.player-1].total_experience / results[sh.data.player-1].total_exp_mass;
                AddToExperience(sh.isPlanet() ? PlanetExperience(sh.data.id) : ShipExperience(sh.data.id), new_exp);
                if (verbose_level > 2)
                    Info("      Unit %d gets %d EP", i, new_exp);
            }
        }
    }
    for (int i = 1; i <= NUM_OWNERS; ++i) {
        int pts = results[i-1].total_pbps_times_1000 / 1000;
        if (pts) {
            if (verbose_level > 2)
                Info("      Player %d gets %d PAL points", i, pts);
#ifdef NEW_PDK
            AddToActivityLevel(i, pts);
#else
            SetActivityLevel(i, GetActivityLevel(i) + pts);
#endif
        }
    }
}

/** Send combat summary to all players. */
static void
sendSummaryMessage(FlakBattle& battle, int player, int x, int y)
{
    string_t text = format(translateMessage("(-h0000)<<< Combat Report >>>\n"
                                            "\n"
                                            "A battle involving %d units\n"
                                            "happened at (%d,%d).\n"
                                            "\n"
                                            "Attendance List:\n", player),
                           battle.getNumShips(), x, y);
    int lines = 6;
    for (int i = 0; i < battle.getNumShips(); ++i) {
        if (lines > 17) {
            doMessage(player, text);
            text = translateMessage("(-h0000)<<< Combat Report >>>\n"
                                    "\n"
                                    "Continued battle report:\n", player);
            lines = 3;
        }

        FlakShip& sh = battle.getShipByNumber(i);
        if (sh.isPlanet())
            text += format(translateMessage("pl%d: %s (player %d)\n", player), sh.data.id, CSTR(sh.data.name), sh.data.player);
        else
            text += format(translateMessage("sh%d: %s (player %d)\n", player), sh.data.id, CSTR(sh.data.name), sh.data.player);
        if (sh.data.ending_status == 0)
            text += translateMessage("  (survived)\n", player);
        else if (sh.data.ending_status == -1)
            text += translateMessage("  (destroyed)\n", player);
        else if (sh.data.ending_status == sh.data.player)
            text += translateMessage("  (captured back)\n", player);
        else
            text += format(translateMessage("  (captured by player %d)\n", player), sh.data.ending_status);

        lines += 2;
    }

    doMessage(player, text);
}

static void
safeRead(std::FILE* file, void* buffer, std::size_t size, const char name[])
{
    if (std::fread(buffer, 1, size, file) != size)
        ErrorExit("Error reading `%s'", name);
}

static void
safeWrite(std::FILE* file, const void* buffer, std::size_t size, const char name[])
{
    if (std::fwrite(buffer, 1, size, file) != size)
        ErrorExit("Error writing `%s'", name);
}

/** Compute ambient flags for some position. I want to use these later
    to add some spice to the visualisation. That is, if there is a
    star cluster, a huge web field or an ion storm, the player should
    display something like that. It's just a bitfield, we don't want
    to give out too much information of course. */
static int32
getAmbientFlags(int x, int y)
{
    // FIXME
    (void) x;
    (void) y;

    return 0;
}

/** Check whether the given owner is valid. We must handle (and
    ignore) out-of-range values because they would crash the player. */
static inline bool
isValidOwner(int owner)
{
    return (owner > 0 && owner <= NUM_OWNERS);
}

/** Generate one combat event.
    \param ship_ids    list of ships
    \param count       number of ships in /ship_ids/
    \param with_planet when enabled, include the planet */
static void
doCombat(const int ship_ids[], size_t count, bool with_planet)
{
    ASSERT(count > 0);

    int x = ShipLocationX(ship_ids[0]), y = ShipLocationY(ship_ids[0]);
    int planet_id = FindPlanetAtShip(ship_ids[0]);
    if (ragnarok_flag || !with_planet)
        planet_id = 0;

    std::vector<ShipInfo> ships;
    for (size_t i = 0; i < count; ++i)
        if (isValidOwner(ShipOwner(ship_ids[i])))
            ships.push_back(ShipInfo().initFromShip(ship_ids[i]));
    if (planet_id)
        if (isValidOwner(PlanetOwner(planet_id)))
            ships.push_back(ShipInfo().initFromPlanet(planet_id));

    if (ships.size() < 2)
        return;

    /* Now we have a vector containing 'ships'. We will now sort it by
       player, then by battle order; and then build up fleets. In
       order to avoid that we always get the same order, first permute
       players. */
    std::sort(ships.begin(), ships.end(), PermutedSorter());

    /* Count players */
    SmallSet<int> players;
    int player_count = 0;
    for (size_t i = 0; i < ships.size(); ++i) {
        if (!players.contains(ships[i].data.player)) {
            players |= ships[i].data.player;
            ++player_count;
        }
    }

    if (player_count < 2)
        return;

    if (verbose_level > 1)
        Info("  Meeting of %d players with %d units at %d,%d",
             player_count, int(ships.size()), x, y);

    /* Now build fleets. */
    FlakBattle battle;
    int  cur_player = 0;
    int  cur_fcbo   = 0;
    bool cur_planet = false;
    int  cur_fleet  = 0;
    int  cur_fleet_size = 0;
    for (FlakBattle::ship_t i = 0; i < ships.size(); ++i) {
        if (i == 0 || cur_player != ships[i].data.player
            || cur_fcbo != ships[i].fcbo_plus_100 / 100
            || cur_planet != ships[i].isPlanet()
            || cur_fleet_size >= flak_config.MaximumFleetSize)
        {
            cur_fleet      = battle.addFleet(ships[i].data.player);
            cur_fcbo       = ships[i].fcbo_plus_100 / 100;
            cur_planet     = ships[i].isPlanet();
            cur_player     = ships[i].data.player;
            cur_fleet_size = 0;
        }
        FlakBattle::ship_t sid = battle.addShip(ships[i].data);
        ASSERT(sid == i);
        ++cur_fleet_size;
    }

    if (verbose_level > 1)
        Info("    Found %d fleets", battle.getNumFleets());
    ASSERT(battle.getNumFleets() > 1);

    /* Now we have all the fleets, compute attack lists */
    for (int i = 0; i < battle.getNumFleets(); ++i)
        computeAttackList(battle, battle.getFleetByNumber(i));

    /* Compute speeds, etc. */
    battle.initAfterSetup();

    /* set random seed */
    {
        /* PDK's `RandomRange' only supplies less than 16 bits. We
           want full 32 bit seeds. This is a cruel hack, but it's
           portable and deterministic. We could as well refrain from
           using the PDK's RNG altogether. However, if we'd use the
           same RNG for playback and initialisation, we would leak too
           much information. For example, if we know the seed of fight
           #1, we could just advance through all numbers and finally
           know the seeds used to set up fight #2 (e.g. whether the
           `+20' bonus comes from a dice or PE). Likewise, we could
           (sort-of) reconstruct battles we do not see. To cut a long
           story short, we definitely need two RNGs, and using PDK's
           doesn't hurt. However, this *does* effectively limit us to
           64k battles (not 4G). */
        uint32 seed = RandomRange(256);
        seed |= RandomRange(256) * 256UL;
        seed |= RandomRange(256) * 65536UL;
        seed |= RandomRange(256) * 16777216UL;
        battle.setInitialSeed(seed);
    }

    /* Empty fight? The fight is either empty, or contains at least
       two fleets. If there were only one fleet, it would not have an
       opponent and so would have been deleted, too. */
    if (battle.getNumFleets() == 0) {
        if (verbose_level > 1)
            Info("    No aggression, skipping this one");
        return;
    }
    ASSERT(battle.getNumFleets() > 1);

    ++battle_counter;
    if (verbose_level > 1) {
        Info("    Found %d aggressive fleets with %d units", battle.getNumFleets(), battle.getNumShips());
        if (verbose_level > 2) {
            for (int i = 0; i < battle.getNumFleets(); ++i)
                Info("      Fleet %d has %d attack list entries",
                     i, battle.getFleetByNumber(i).data.att_list_size);
            for (int i = 0; i < battle.getNumShips(); ++i) {
                Info("      Unit %d: player %d, mass %d, name '%s'", i,
                     battle.getShipByNumber(i).data.player,
                     battle.getShipByNumber(i).data.mass,
                     CSTR(battle.getShipByNumber(i).data.name));
                if (battle.getShipByNumber(i).isPlanet())
                    Info("        (a planet with %d DP)", PlanetDefense(battle.getShipByNumber(i).data.id));
                else
                    Info("        (a '%s' ship)", CSTR(HullName(battle.getShipByNumber(i).data.hull, 0)));
            }
        }
    } else if (verbose_level > 0) {
        Info("  Fight %d (%d units in %d fleets)...", battle_counter,
             battle.getNumShips(), battle.getNumFleets());
    }

    /* up we go... */
    FlakNullVisualizer nv(battle);
    battle.init();
    while (!battle.playOneTick())
        ;
    if (verbose_level > 1)
        Info("    Fight took %ld ticks to resolve", (long) battle.getTime());

    /* resolve it */
    resolveBattle(battle);

    /* send messages */
    bool old_display_message_flag = display_messages_flag;
    for (int pl = 1; pl <= FLAK_NUM_OWNERS; ++pl) {
        if (players.contains(pl)) {
            sendSummaryMessage(battle, pl, x, y);
            display_messages_flag = false;
        }
    }
    display_messages_flag = old_display_message_flag;

    /* write it out */
    Buffer<char> tmp;
    battle.saveToStream(tmp, x, y, getAmbientFlags(x, y));
    safeWrite(data_file, tmp.ptr(), tmp.getSize(), DATA_FILE_NAME);

    /* finally, remember who sees it. This is not nice, but it will do. */
    SmallSet<int> who_sees_it;
    players = SmallSet<int>();
    for (int i = 0; i < battle.getNumShips(); ++i)
        players += battle.getShipByNumber(i).data.player;
    for (int i = 1; i <= RACE_NR; ++i) {
        if (players.contains(i)) {
            who_sees_it += i;
            for (int j = 1; j <= RACE_NR; ++j) {
                if (j != i && PlayersAreAllies(RaceType_Def(i), RaceType_Def(j))
                    && PlayerAllowsAlly(RaceType_Def(i), RaceType_Def(j), ALLY_VISION)
                    && PlayerAllowsAlly(RaceType_Def(i), RaceType_Def(j), ALLY_COMBAT)
                    && PlayerAllowsAlly(RaceType_Def(i), RaceType_Def(j), ALLY_SHIPS))
                {
                    who_sees_it += j;
                }
            }
        }
    }
    battle_vis_info.push_back(who_sees_it);
}

/** Do all normal combat. */
static void
doCombat()
{
    /* build array of ships in X,Y order */
    std::vector<int> ships;
    for (int i = 1; i <= SHIP_NR; ++i)
        if (IsShipExist(i))
            ships.push_back(i);
    if (ships.empty())
        return;

    std::sort(ships.begin(), ships.end(), sortByLocationPredicate);

    Info("Normal combat order...");

    std::size_t first = 0;
    if (!ragnarok_flag) {
        for (std::size_t i = 1; i < ships.size(); ++i) {
            if (ShipLocationX(ships[i]) != ShipLocationX(ships[first]) || ShipLocationY(ships[i]) != ShipLocationY(ships[first])) {
                doCombat(&ships[first], i - first, true);
                first = i;
            }
        }
    }
    doCombat(&ships[first], ships.size() - first, true);
}

/** Do intercept attack. */
static void
doInterceptAttack()
{
    enum { eIsInterceptor = 1, eIsTarget = 2, eDoNow = 4 };

    std::vector<int> ship_info(SHIP_NR+1);
    for (int i = 1; i <= SHIP_NR; ++i) {
        if (IsShipExist(i)
            && ShipMission(i) == Intercept
            && ShipCargo(i, NEUTRONIUM) != 0
            && ShipCanCloak(i))
        {
            int target = ShipInterceptTarget(i);
            char ifc[4], tfc[4];
            if (target > 0 && target <= SHIP_NR
                && target != i
                && IsShipExist(target)
                && !IsShipCloaked(target)
                && ShipCargo(target, NEUTRONIUM) != 0
                && ShipOwner(i) != ShipOwner(target)
                && ShipEnemy(i) == ShipOwner(target)
                && ! (PlayersAreAllies(ShipOwner(i), ShipOwner(target))
                      && PlayerAllowsAlly(ShipOwner(i), ShipOwner(target), ALLY_COMBAT))
                && ShipLocationX(i) == ShipLocationX(target)
                && ShipLocationY(i) == ShipLocationY(target)
                && !IsFCodeMatch(ShipFCode(i, ifc), ShipFCode(target, tfc)))
            {
                ship_info[i] |= eIsInterceptor;
                ship_info[target] |= eIsTarget;
            }
        }
    }

    Info("Intercept attacks...");

    /* We have some relation 'a < b' meaning 'a intercept-attacks b'.
       Every *valid* a is marked as eIsInterceptor in ship_info, and
       every corresponding b is marked as eIsTarget.

       For every 'a', there is at most one 'b' such that 'a < b', but
       for every 'b' there can be many 'a', so we essentially have a
       tree-like structure.

       We therefore just locate any random member 'start' and walk
       upwards towards the root of the tree, getting an initial set
       'B'. We can now repeatedly find elements 'c' such that 'c < b,
       b \in B', and add them to B. */
    while (1) {
        /* find some interceptor */
        int start = 0;
        for (int i = 1; i <= SHIP_NR; ++i) {
            if (ship_info[i] & eIsInterceptor) {
                start = i;
                break;
            }
        }
        if (!start)
            break;              // end this phase

        /* okay, we found one. Mark everything forward. Be careful with loops. */
        while (1) {
            ship_info[start] |= eDoNow;
            if (!(ship_info[start] & eIsInterceptor))
                break;
            start = ShipInterceptTarget(start);
            if (ship_info[start] & eDoNow)
                break;
        }

        /* mark everything else (transitive closure) */
        while (1) {
            bool done = false;
            for (int i = 1; i <= SHIP_NR; ++i) {
                if ((ship_info[i] & eIsInterceptor)
                    && !(ship_info[i] & eDoNow)
                    && (ship_info[ShipInterceptTarget(i)] & eDoNow))
                {
                    ship_info[i] |= eDoNow;
                    done = true;
                }
            }
            if (!done)
                break;
        }

        /* build list of ids, and unmark ships */
        std::vector<int> ship_ids;
        for (int i = 1; i <= SHIP_NR; ++i) {
            if (ship_info[i] & eDoNow) {
                ship_info[i] = 0;
                ship_ids.push_back(i);
            }
        }

        /* combat */
        doCombat(&*ship_ids.begin(), ship_ids.size(), false);
    }
}

/** Write header to data_file. */
static void
writeHeader()
{
    using namespace std;
    TFlakHeader h;
    memcpy(h.magic, FLAK_MAGIC, sizeof(h.magic));
    h.filefmt_version = 0;
    h.player          = 0;
    h.turn            = TurnNumber();
    h.num_battles     = battle_counter;
    HostTimeStamp(h.timestamp);
    h.reserved = 0;

    char buffer[h.size];
    storeStructure(buffer, h);
    safeWrite(data_file, buffer, sizeof(buffer), DATA_FILE_NAME);
}

/** Generate flakX.dat file for one player. */
static void
generatePlayerFile(int player)
{
    TFlakHeader header;
    char fn_buffer[20], hdr_buffer[header.size];
    int num_in, num_out;

    sprintf(fn_buffer, "flak%i.dat", player);
    std::FILE* in  = OpenInputFile(DATA_FILE_NAME, GAME_DIR_ONLY);
    std::FILE* out = OpenOutputFile(fn_buffer, GAME_DIR_ONLY);

    safeRead(in, hdr_buffer, sizeof(hdr_buffer), DATA_FILE_NAME);
    getStructure(hdr_buffer, header);
    num_in = header.num_battles;
    num_out = 0;

    safeWrite(out, hdr_buffer, sizeof(hdr_buffer), fn_buffer);

    for (int i = 0; i < num_in; ++i) {
        Buffer<char> ch;
        ch.ensureSize(4);
        safeRead(in, ch.ptr(), 4, DATA_FILE_NAME);
        ch.ensureSize(getUint32(ch.ptr()));
        safeRead(in, ch.ptr()+4, ch.getSize()-4, DATA_FILE_NAME);
        if (!battle_vis_info[i].contains(player))
            continue;
        safeWrite(out, ch.ptr(), ch.getSize(), fn_buffer);
        ++num_out;
    }

    header.num_battles = num_out;
    header.player      = player;
    storeStructure(hdr_buffer, header);
    std::fseek(out, 0, SEEK_SET);
    safeWrite(out, hdr_buffer, sizeof(hdr_buffer), fn_buffer);

    std::fclose(in);
    std::fclose(out);

    if (verbose_level > 1)
        Info("Wrote %d battles for player %d", num_out, player);
}

/** Main entry point. */
int
main(int argc, char** argv)
{
    int n;
    program_basename = argv[0];
    while (const char* p = std::strpbrk(program_basename, ":/\\"))
        program_basename = p+1;
    while ((n = pgetopt(argc, argv, const_cast<char*>("hvVarmZ"))) != EOF) {
        if (n == 'h') {
            help();
        } else if (n == 'v') {
            version();
        } else if (n == 'V') {
            ++verbose_level;
        } else if (n == 'a') {
            aggression_flag = true;
        } else if (n == 'r') {
            readonly_flag = true;
        } else if (n == 'm') {
            display_messages_flag = true;
        } else if (n == 'Z') {
            ragnarok_flag = true;
        } else {
            help();
        }
    }

    if (argv[poptind])
        gGameDirectory = argv[poptind++];
    if (argv[poptind])
        gRootDirectory = argv[poptind++];
    if (argv[poptind])
        ErrorExit("Too many parameters. Try `%s -h' for help.", argv[0]);

    enableStdioConsole();
    InitPHOSTLib();
    if (!ReadGlobalData())
        ErrorExit("Unable to read global data.");
    if (!ReadHostData())
        ErrorExit("Unable to read host data.");

    loadFlakConfig();
    initTranslation();

    data_file = OpenOutputFile(DATA_FILE_NAME, GAME_DIR_ONLY);
    battle_counter = 0;
    writeHeader();
    if (gPconfigInfo->AllowInterceptAttack && !ragnarok_flag)
        doInterceptAttack();
    doCombat();
    std::fseek(data_file, 0, SEEK_SET);
    writeHeader();
    std::fclose(data_file);

    for (int i = 1; i <= RACE_NR; ++i)
        if (PlayerIsActive(RaceType_Def(i)))
            generatePlayerFile(i);

    if (!readonly_flag)
        WriteHostData();
    FreePHOSTLib();
}
