/**
  *  \file io/glob.cc
  *  \brief Globbing
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include "io/glob.h"
#include "util/nls.h"
#include "cpluslib/assert.h"

/*! \class FileNamePatternException
    \brief Malformed File Name Pattern

    Thrown by globbing functions when a glob pattern is invalid (e.g. 
    ends with a quote character).  */

/** Construct from string. */
FileNamePatternException::FileNamePatternException(const string_t& awhat)
    : UserException(awhat)
{ }

/** Construct from C string. */
FileNamePatternException::FileNamePatternException(const char* awhat)
    : UserException(awhat)
{ }

/** Copy constructor. */
FileNamePatternException::FileNamePatternException(const FileNamePatternException& other)
    : UserException(other)
{ }

FileNamePatternException::~FileNamePatternException() throw()
{ }


/*! \class FileNameMatcher
    \brief File Name Pattern Matcher

    This class provides a means of compiling and applying a file name
    pattern. This default implementation provides the following rules:
    - "*" matches any sequence of characters, including nothing at all
    - "?" matches any single character
    - "\" quotes the next character, i.e. "\*" matches a single star

    This class is the internal back-end to FileNamePattern.

    \todo We should optionally use system rules, i.e. fnmatch(3) on
    Unix, respective Windows function on Windows. This requires a new
    function under arch/, and a configure switch (maybe users prefer
    uniform behaviour on all platforms?). This default implementation
    does not implement any operating system's rules completely.

    \todo Right now, this isn't particularily efficient. We use a
    simple back-tracking matcher. */
class FileNameMatcher {
    enum OpType {
        OpMatchLiteral,         ///< Match the literal specified in the string argument.
        OpMatchAnyChar,         ///< Match exactly one arbitrary character.
        OpMatchAnyString        ///< Match any number, including zero, arbitrary characters.
    };
    struct Operation {
        OpType   type;
        string_t arg;

        Operation(OpType type, string_t arg)
            : type(type), arg(arg)
            { }
    };
    std::vector<Operation> ops;

    void addOp(OpType op, const string_t arg);
    bool matchRecursive(const string_t& arg,
                        string_t::size_type arg_pos,
                        std::vector<Operation>::size_type match_pos) const;
 public:
    FileNameMatcher();
    ~FileNameMatcher();

    void compile(string_t arg);
    bool match(string_t arg) const;
    bool hasWildcard() const;
    string_t getFileName() const;
};

/** Construct blank object. */
FileNameMatcher::FileNameMatcher()
{ }

/** Destroy. */
FileNameMatcher::~FileNameMatcher()
{ }

/** Add operation. This does a few trivial optimisations. The goal is
    to speed up matching ("***x" takes cubic time without
    optimisation, linear time with), and to otherwise simplify
    handling. */
void
FileNameMatcher::addOp(OpType op, const string_t arg)
{
    /* attempt to optimize */
    if (ops.size() != 0) {
        Operation& last_op = ops.back();
        switch (op) {
         case OpMatchLiteral:
            if (last_op.type == OpMatchLiteral) {
                /* two literals, as in "foo\*", compile into just one
                   check */
                last_op.arg += arg;
                return;
            }
            break;
         case OpMatchAnyString:
            if (last_op.type == OpMatchAnyString) {
                /* two stars, as in "foo**", equivalent to just "foo*" */
                return;
            }
            break;
         case OpMatchAnyChar:
            if (last_op.type == OpMatchAnyString) {
                /* optimise "*?" into "?*" */
                last_op.type = OpMatchAnyChar;
                ops.push_back(Operation(OpMatchAnyString, string_t()));
                return;
            }
            break;
         default:
            break;
        }
    }

    /* add it */
    ops.push_back(Operation(op, arg));
}

/** Compile pattern for later use. This builds an internal private
    representation which simplifies and speeds up later matching.
    \throw FileNamePatternException if the pattern is syntactically
    invalid */
void
FileNameMatcher::compile(const string_t arg)
{
    ops.clear();

    string_t::size_type pos = 0;
    while (pos < arg.size()) {
        if (arg[pos] == '*') {
            addOp(OpMatchAnyString, string_t());
            ++pos;
        } else if (arg[pos] == '?') {
            addOp(OpMatchAnyChar, string_t());
            ++pos;
        } else if (arg[pos] == '\\') {
            ++pos;
            if (pos >= arg.size())
                throw FileNamePatternException(_("Backslash must be followed by a character"));
            addOp(OpMatchLiteral, string_t(1, arg[pos]));
            ++pos;
        } else {
            addOp(OpMatchLiteral, string_t(1, arg[pos]));
            ++pos;
        }
    }
}

/** Check whether we have a wildcard.
    \return true if this is a wildcard, false if it is a proper file name. */
bool
FileNameMatcher::hasWildcard() const
{
    // Due to optimisation, this is very simple:
    return ops.size() != 1
        || ops.front().type != OpMatchLiteral;
}

/** Get file name.
    \pre !hasWildcard() */
string_t
FileNameMatcher::getFileName() const
{
    ASSERT(!hasWildcard());
    return ops.front().arg;
}

/** Match pattern.
    \param arg file name
    \return true iff arg matches the compiled pattern */
bool
FileNameMatcher::match(string_t arg) const
{
    return matchRecursive(arg, 0, 0);
}

/** Recursive matching. Check whether tail of input string matches
    tail of pattern.
    \param arg       [in] String to match
    \param arg_pos   [in] Position in string to start matching at
    \param match_pos [in] Position in pattern to start matching at
    \return true iff match ok

    \todo We could recognize some common patterns to speed up
    matching, e.g. "*x" = directly compare string end, "*x*" = use
    a single s.find() call. */
bool
FileNameMatcher::matchRecursive(const string_t& arg, string_t::size_type arg_pos,
                                std::vector<Operation>::size_type match_pos) const
{
    while (match_pos < ops.size()) {
        ASSERT(arg_pos <= arg.size());
        const Operation& me = ops[match_pos];
        switch (me.type) {
         case OpMatchLiteral: {
            string_t::size_type arg_avail = arg.size() - arg_pos;
            if (arg_avail < me.arg.size())
                /* pattern longer than string */
                return false;
            if (strCaseCmp(arg.data() + arg_pos, me.arg.size(),
                           me.arg.data(),        me.arg.size()) != 0)
                /* pattern does not match string */
                return false;
            /* success */
            arg_pos += me.arg.size();
            ++match_pos;
            break;
         }
         case OpMatchAnyChar: {
            if (arg_pos >= arg.size())
                /* pattern longer than string */
                return false;
            ++arg_pos;
            ++match_pos;
            break;
         }
         case OpMatchAnyString: {
            /* short-cut: "foo*" */
            if (match_pos + 1 == ops.size())
                return true;
            /* simple, slow way */
            /* FIXME: possible optimisations:
               - optimise "*foo*" to use strstr / find / ...
               - optimise "*foo" to just look at the end of the text */
            do {
                if (matchRecursive(arg, arg_pos, match_pos+1))
                    return true;
                ++arg_pos;
            } while (arg_pos < arg.size());
            return false;
         }
        }
    }

    return arg_pos == arg.size();
}

/**************************** FileNamePattern ****************************/

/*! \class FileNamePattern
    \brief File Name Matching

    This class provides a way to match file names against patterns. 
    You have to place a pattern in a FileNamePattern object and can
    then use it to match against possibly many file names.

    FileNamePattern is the portable interface, it uses an instance of
    FileNameMatcher to implement actual matching rules. */

/** Construct blank pattern. Call setPattern() before use. */
FileNamePattern::FileNamePattern()
    : rep(new FileNameMatcher())
{ }

/** Construct pattern.
    \param name [in] Pattern
    \see FileNameMatcher */
FileNamePattern::FileNamePattern(const string_t name)
    : rep(new FileNameMatcher())
{
    rep->compile(name);
}

/** Copy constructor. */
FileNamePattern::FileNamePattern(const FileNamePattern& other)
    : rep(new FileNameMatcher(*other.rep))
{ }

/** Assignment operator. */
FileNamePattern&
FileNamePattern::operator=(const FileNamePattern& other)
{
    *rep = *other.rep;
    return *this;
}

/** Destructor. */
FileNamePattern::~FileNamePattern()
{
    delete rep;
}

/** Set pattern. Discards the old one and sets a new one.
    \param pattern [in] The new pattern */
void
FileNamePattern::setPattern(const string_t pattern)
{
    rep->compile(pattern);
}

/** Check whether we have a wildcard. If it is not a wildcard, it is a
    proper file name which can be used directly (see getFileName()).
    \return true if this is a wildcard, false if it is a proper file name. */
bool
FileNamePattern::hasWildcard() const
{
    return rep->hasWildcard();
}

/** Get file name. Note that this file name can differ from the name
    used with the constructor in case quoting is used. For example,
    with "foo\\*bar", hasWildcard() is true and getFileName() returns
    "foo*bar"; the backslash quotes.
    \pre hasWildcard() */
string_t
FileNamePattern::getFileName() const
{
    return rep->getFileName();
}

/** Check for match.
    \param filename [in] file name to test
    \return true if filename matches pattern. */
bool
FileNamePattern::match(const string_t filename) const
{
    return rep->match(filename);
}

/** Get pattern that matches a single file.
    \param name [in] The name to match
    \post match(x) <=> x and \c name name the same file */
string_t
FileNamePattern::getSingleFilePattern(string_t name)
{
    string_t::size_type n = 0;
    while ((n = name.find_first_of("\\?*", n)) != name.npos) {
        name.insert(n, "\\");
        n += 2;
    }
    return name;
}

/** Get pattern that matches all files.
    \post match(x) <=> x names a user-visible file */
string_t
FileNamePattern::getAllFilesPattern()
{
    return string_t("*");
}

/** Get pattern that matches all files with a particular extension.
    \param ext [in] Extension, not including the dot */
string_t
FileNamePattern::getAllFilesWithExtensionPattern(string_t ext)
{
    return string_t("*.") + getSingleFilePattern(ext);
}
